#include <process.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define INCL_DOSDEVICES
#define INCL_DOSDEVIOCTL
#define INCL_DOSERRORS
#define INCL_DOSFILEMGR
#define INCL_DOSMODULEMGR
#define INCL_DOSPROCESS
#define INCL_DOSSEMAPHORES
#define INCL_WINFRAMEMGR
#define INCL_WINHOOKS
#define INCL_WININPUT
#define INCL_WINPOINTERS
#define INCL_WINPROGRAMLIST
#define INCL_WINSWITCHLIST
#define INCL_WINSYS
#define INCL_WINTIMER
#define INCL_WINWINDOWMGR
#include <os2.h>

#include "standby2.h"
#include "standby2rc.h"

#define VPCTitle "Virtual PC"

void _Optlink term( int sig );
void _Optlink cont( void *arg );
MRESULT EXPENTRY dialogproc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 );
static void powerondisplay( void );
static void poweroffdisplay( void );
static HAB hab = NULLHANDLE;
static HEV hev = NULLHANDLE;
static PFN pfn = NULL;
static HMODULE hmod = NULLHANDLE;
static HFILE apmdev = NULLHANDLE;
static int powerstate = 1;
static int waitsec = 0;
static ULONG waitmsec = 0;
static int forceoffsec = 0;
static BOOL disablepoweroff = FALSE;
static HWND prevfocushwnd = NULLHANDLE;

#pragma pack (1)
struct ApmParam {
   SHORT subid;
   SHORT reserved;
   SHORT devid;
   SHORT pwrst;
};
#pragma pack ()

main(
int argc,
char *argv[],
char *envp[]) {
   LONG waitmin = 5;
   HMQ hmq = NULLHANDLE;
   ULONG action = 0;
   APIRET rc = 0;
   HWND hwnd = NULLHANDLE;
   QMSG qmsg;                 
   CHAR loaderror[BUFSIZ];
   SWCNTRL swcntrl;                        
   HSWITCH hswitch = NULLHANDLE;
   HPOINTER hicon = NULLHANDLE;
   LONG cx = 0;
   LONG cy = 0;
   SWP swp;
   if ( argc > 2 ) {
      exit(8);
   } /* endif */
   if ( 2 == argc ) {
      waitmin = atoi(argv[1]);
   } /* endif */
   if ( waitmin <= 0 ) {
      exit(8);
   } /* endif */
   waitsec = 60 * waitmin;
   waitmsec = 1000 * waitsec;
   action = 0;
   rc = DosOpen( APMDEV,
                 &apmdev,
                 &action,
                 0,
                 0,
                 OPEN_ACTION_OPEN_IF_EXISTS,
                 OPEN_SHARE_DENYNONE,
                 0 );
   if (rc != NO_ERROR) {
      exit(8);
   } /* endif */
   strcpy( loaderror, "" );
   rc = DosLoadModule(loaderror, sizeof(loaderror), "STANDBD2", &hmod);
   if (NO_ERROR != rc) {
      exit(8);
   } /* endif */
   pfn = NULL;
   rc = DosQueryProcAddr(hmod, 0, "InputHook", &pfn);
   if (NO_ERROR != rc) {
      exit(8);
   } /* endif */     
   hab = WinInitialize( 0 );
   if (!hab) {
      exit(8);
   } /* endif */
   rc = DosCreateEventSem( SemName, &hev, DC_SEM_SHARED, FALSE );
   if ( NO_ERROR != rc ) {
      if (rc != ERROR_DUPLICATE_NAME ) {
         exit(8);
      } /* endif */
      rc = DosOpenEventSem( SemName, &hev );
      if ( NO_ERROR != rc ) {
         exit(8);
      } /* endif */     
   } /* endif */          
   if (!WinSetHook(hab, 0, HK_INPUT, pfn, hmod)) {
      exit(8);
   } /* endif */
   signal( SIGBREAK, term );
   signal( SIGTERM, term );
   hmq = WinCreateMsgQueue( hab, 0 );
   if ( !hmq ) {
      exit(8);
   } /* endif */
   hwnd = WinLoadDlg( HWND_DESKTOP,
                      HWND_DESKTOP,
                      (PFNWP)dialogproc,
                      (HMODULE)0,
                      DID_PRODINFO,
                      NULL);
   if ( !hwnd ) {
      exit(8);
   } /* endif */
   cx = WinQuerySysValue( HWND_DESKTOP, SV_CXFULLSCREEN );
   cy = WinQuerySysValue( HWND_DESKTOP, SV_CYFULLSCREEN );
   if ( cx && cy && WinQueryWindowPos( hwnd, &swp ) ) {
      WinSetWindowPos( hwnd, NULLHANDLE, (cx - swp.x) / 2, (cy - swp.y) / 2, 0, 0, SWP_MOVE );
   } /* endif */
   memset( &swcntrl, 0, sizeof(swcntrl) );
   swcntrl.hwnd = hwnd;
   swcntrl.idProcess = getpid();
   swcntrl.uchVisibility = SWL_VISIBLE;
   swcntrl.fbJump = SWL_JUMPABLE;
   strcpy( swcntrl.szSwtitle, "Standby2" );
   swcntrl.bProgType = PROG_DEFAULT;
   if ( !WinCreateSwitchEntry( hab, &swcntrl ) ) {
      exit(8);
   } /* endif */
   hicon = WinLoadPointer( HWND_DESKTOP, NULLHANDLE, 1 );
   if ( hicon ) {
      WinSendMsg( hwnd, WM_SETICON, (MPARAM)hicon, 0 );
   } /* endif */
   _beginthread( cont, NULL, 0x10000, NULL );
   while ( WinGetMsg( hab, &qmsg, 0, 0, 0) ) {
      WinDispatchMsg( hab, &qmsg );
   } /* endwhile */
   WinDestroyPointer( hicon );
   WinDestroyWindow( hwnd );
   WinDestroyMsgQueue( hmq );
   rc = DosCloseEventSem( hev );
   WinReleaseHook(hab, 0, HK_INPUT, pfn, hmod);
   rc = DosFreeModule( hmod );
   WinTerminate( hab );
   rc = DosClose( apmdev );
   powerondisplay();
   return (0);
}

void _Optlink
term( 
int sig ) {
   APIRET rc = DosCloseEventSem( hev );
   WinReleaseHook(hab, 0, HK_INPUT, pfn, hmod);
   WinTerminate( hab );
   powerondisplay();
   rc = DosClose( apmdev );
}

static void
powerondisplay() {
   struct ApmParam param;
   ULONG szparam = 0;
   USHORT data = 0;
   ULONG szdata = 0;
   APIRET rc = 0;
   param.subid = SetPowerState;
   param.reserved = 0;
   param.devid = AllDisplays;
   param.pwrst = Ready;
   szparam = sizeof(param);
   data = 0;
   szdata = sizeof(data);
   rc = DosDevIOCtl( apmdev,
                     0x0c,
                     0x40,
                     &param,
                     sizeof(param),
                     &szparam,
                     &data,
                     sizeof(data),
                     &szdata);
}

static 
void poweroffdisplay( void ) {
   struct ApmParam param;
   ULONG szparam = 0;
   USHORT data = 0;
   ULONG szdata = 0;
   APIRET rc = 0;
   param.subid = SetPowerState;
   param.reserved = 0;
   param.devid = AllDisplays;
   param.pwrst = Standby;
   szparam = sizeof(param);
   data = 0;
   szdata = sizeof(data);
   rc = DosDevIOCtl( apmdev,
                     0x0c,
                     0x40,
                     &param,
                     sizeof(param),
                     &szparam,
                     &data,
                     sizeof(data),
                     &szdata);
}

void _Optlink 
cont( void *arg ) {
   for (; ; ) {
      ULONG count = 0;
      APIRET rc = DosResetEventSem( hev, &count );
      if ( rc != NO_ERROR && rc != ERROR_ALREADY_RESET ) {
         raise( SIGTERM );
      } /* endif */
      rc = DosWaitEventSem( hev, waitmsec );
      if ( NO_ERROR == rc ) {
         if ( time(NULL) < forceoffsec ) {
            continue;
         } /* endif */
         waitmsec = 1000 * waitsec;
         if ( powerstate ) {
            continue;
         } /* endif */
         if ( prevfocushwnd ) {
             WinSetFocus( HWND_DESKTOP, prevfocushwnd );
         } /* endif */
         powerondisplay();
         powerstate = 1;
      } else if ( ERROR_TIMEOUT == rc ) {
         if ( disablepoweroff ) {
            continue;
         } /* endif */
         waitmsec = SEM_INDEFINITE_WAIT;
         if ( !powerstate ) {
            continue;
         } /* endif */
         prevfocushwnd = WinQueryFocus( HWND_DESKTOP );
         WinSetFocus( HWND_DESKTOP, HWND_DESKTOP );
         poweroffdisplay();
         powerstate = 0;
      } /* endif */
   } /* endfor */
}

MRESULT EXPENTRY 
dialogproc( 
HWND hwnd, 
ULONG msg, 
MPARAM mp1, 
MPARAM mp2 ) {
   HWND buttonhwnd = NULLHANDLE;
   switch ( msg ) {
   case WM_COMMAND:
      switch ( SHORT1FROMMP(mp1) ) {
      case DID_QUIT:
         WinPostMsg( hwnd, WM_QUIT, 0, 0 );
         break;
      case DID_DISABLE:
         buttonhwnd = WinWindowFromID(hwnd, DID_DISABLE);
         if ( !buttonhwnd ) {
            exit( 8 );
         } /* endif */
         if ( disablepoweroff ) {
            disablepoweroff = FALSE;
            WinSetWindowText( buttonhwnd, "Disable" );
         } else {
            disablepoweroff = TRUE;
            WinSetWindowText( buttonhwnd, "Enable" );
         } /* endif */
         WinSetWindowPos( hwnd, NULLHANDLE, 0, 0, 0, 0, SWP_HIDE );
         break;
      case DID_OK:
         WinSetWindowPos( hwnd, NULLHANDLE, 0, 0, 0, 0, SWP_HIDE );
         break;
      case DID_STANDBY:
         WinSetWindowPos( hwnd, NULLHANDLE, 0, 0, 0, 0, SWP_HIDE );
         forceoffsec = time( NULL ) + 4;
         poweroffdisplay();
         powerstate = 0;
         break;
      } /* endswitch */
      return ( 0 );
   default:
     return WinDefDlgProc( hwnd, msg, mp1, mp2 );
   } /* endswitch */
   return ( 0 );
}

